import java.util.Vector;

/**
 * Classe AllinParziale: calcola l'allineamento ottimo parziale 
 * tra due sequenze di aminoacidi, utilizzando una matrice di sostituzione
 * 
 */
public class AllinParziale extends Allineamento
{
    // instance variables
    private int[][] s; // matrice dei punteggi di similarit
    private byte[][] p; // matrice dei percorsi: da cella x alla precedente 
                        // che possono essere 1, 2, 4, 1+2=3, 1+4=5, 2+4=6, 1+2+4=7
                        //    [2]  [4]  []
                        //    [1]  [x]  []
                        //    []   []   []
    
    /**
     * Constructor for objects of class Allineamento                                                           
     */
    public AllinParziale(Sequenza sequenza1, Sequenza sequenza2, MatriceSostituzione matrice, PenalitGap penalitGap) //int penalitGap)
    {
        // initialise instance variables
        super(sequenza1, sequenza2, matrice, penalitGap);
        //
        int n = sequenza1.lunghezza();
        int m = sequenza2.lunghezza();
        // n+1 e m+1 perch la prima colonna e la prima riga contengono il carattere "gap"
        s = new int[n+1][m+1];  
        p = new byte[n+1][m+1];

    }
    

    /**
     * Metodo che calcola l'allineamento parziale ottimo
     * in sostanza si considera il miglior allineamento considerando una sequenza come sottosequenza dell'altra
     * ==> non penalizza i gap alle estremit
     */
    public void eseguiAllineamento()
    {
        Sequenza a = sequenza1();
        Sequenza b = sequenza2();
        MatriceSostituzione mat = matriceSostituzione();
        PenalitGap pg = penalitGap();
        int n = a.lunghezza();
        int m = b.lunghezza();
        int d = pg.costoCreazione();
        int e = pg.costoEstensione();

        // lascia a 0 le celle della prima colonna e della prima riga
        int i;
        int j;
         
       // ed ora calcolo tutti i punteggi di similarit 
       // secondo l'algoritmo di Neddleman-Wunsch
       int s1;
       int s2;
       int s4;
       int x;
       int y;
       int max;
       int prec;
       i = 1;
       while (i < n+1)
       {
           j = 1; 
           while (j < m+1)
           {
               x = a.indiceAminoacido(i-1);
               y = b.indiceAminoacido(j-1);
               
               if (p[i][j-1] == 1)
                  {s1 = s[i][j-1] - e;}
               else   
                  {s1 = s[i][j-1] - d;}                              
               s2 = s[i-1][j-1] + mat.punteggio(x,y);
               if (p[i-1][j] == 4)
                   {s4 = s[i-1][j] - e;}
               else
                   {s4 = s[i-1][j] - d;}
               // cerco il massimo punteggio
               prec = 1;  // vai a sinistra
               max = s1;
               if (s2 > max)
                  {
                    prec = 2;  // vai in diagonale
                    max = s2;
                  }
               else if (s2 == max)
                  {
                    prec = prec + 2;  // vai anche in diagonale
                  }
               
               if (s4 > max)
                  {
                    prec = 4;  // vai in alto
                    max = s4;
                  }
               else if (s4 == max)
                  {
                    prec = prec + 4;  // vai anche in alto
                  }
                  
               s[i][j] = max;
               p[i][j] = (byte)prec;
               

               j++;
           }
           
           i++;
       }           
    }



     /**
     * Determina il percorso ottimo parziale (dalla matrice p dei percorsi)
     *
     * @return     array di Stringhe con tutti i percorsi ottimi
     */
    public String[] percorsoOttimo()
    {
        Sequenza a = sequenza1();
        Sequenza b = sequenza2();
        MatriceSostituzione mat = matriceSostituzione();
        PenalitGap pg = penalitGap();
        Vector v = new Vector();  // contiene i vari percorsi ottimi
        String percorso = "";  // "124..."  un singolo percorso ottimo
        int n = a.lunghezza();
        int m = b.lunghezza();
        int pmax = maxPunteggio();
        
        // cerca punteggio massimo nell'ultima riga e nell'ultima colonna
        int q = 0; // indice di scorrimento
        while (q < m+1)
           {
             if (s[n][q] == pmax)
                {
                   // chiamata ricorsiva
                   risaliPercorso(n, q, v, percorso);                
                }
             q++;           
           }
        q = 0;
        while (q < n)  // non n+1 per non considerare due volte s[n][m]
           {
             if (s[q][m] == pmax)
                {
                   // chiamata ricorsiva
                   risaliPercorso(q, m, v, percorso);                                
                }
             q++;           
           }
        //  preparo l'array da restituire
        int z = v.size();
        //  uso 5 stringhe per restituire ciascun allineamento
        String[] arr = new String[5*z];  // sequenza a, matches, sequenza b, 2 statistiche
        String s;
        char c;  // l'aminoacido corrente
        int x; // indice per la sequenza a
        int y; // indice per la sequenza b
        // statistiche
        int contaMatch;
        int contaPositivi;
        int lunghPercorso;
        //
        int i = 0;  // indice per l'array di stringhe da restituire
        int k = 0;  // indice del percorso corrente
        int j;  // indice per scorrere il percorso corrente
        while (k < z)
          {
            x = 0;  
            y = 0;  
            arr[i] = "";
            arr[i+1] = "";
            arr[i+2] = "";
            s = (String)v.elementAt(k);
            contaMatch = 0;
            contaPositivi = 0;
            
            lunghPercorso = s.length();
            // non devo considerare i gap iniziali (si tratta di una serie di '1' o di '4' finali!)
            int t = s.length()-1; // indice di scorrimento
            char c1 = s.charAt(t);
            boolean finito;
            if ((c1 == '1') || (c1 == '4'))
               {finito = false;}
            else
               {finito = true;}               
            while (!finito)
            {
                if (s.charAt(t) == c1)
                {
                   lunghPercorso--;
                   t--;
                }
                else
                {finito = true;}
            }  
            // 
            j = s.length() - 1;
            while (j >= 0)
               {                  
                  c = s.charAt(j);
                  if (c == '1')
                     {
                        arr[i] = arr[i] + "-";  // gap
                        arr[i+1] = arr[i+1] + " ";  // mismatch
                        arr[i+2] = arr[i+2] + b.aminoacido(y);  // aminoacido
                        y++;
                     }
                  else if (c == '2')   
                    {
                        arr[i] = arr[i] + a.aminoacido(x);  // aminoacido
                        if (a.aminoacido(x) == b.aminoacido(y))
                           {  // match
                              arr[i+1] = arr[i+1] + a.aminoacido(x);
                              contaMatch++; contaPositivi++;
                            }  
                        else if (mat.punteggio(a.indiceAminoacido(x),  b.indiceAminoacido(y)) > 0)
                           {  // positivo
                              arr[i+1] = arr[i+1] + "+";
                              contaPositivi++;
                           }  
                        else   
                           {arr[i+1] = arr[i+1] + " ";}  // mismatch
                        arr[i+2] = arr[i+2] + b.aminoacido(y);  // aminoacido
                        x++;  y++;
                     }
                  else if (c == '4')
                     {
                        arr[i] = arr[i] + a.aminoacido(x);  // aminoacido
                        arr[i+1] = arr[i+1] + " ";  // mismatch
                        arr[i+2] = arr[i+2] + "-";  // gap
                        x++;
                     }
                  j--;   
               } 
              arr[i+3] = "punteggio =  " + maxPunteggio() + " con " + mat + " " + pg;
              arr[i+4] = "identit = " + contaMatch + "/" +  lunghPercorso + " (" + Math.round((1.0 * contaMatch)/lunghPercorso*100)+"%) ";
              arr[i+4] = arr[i+4] + " positivi = " + contaPositivi + "/" + lunghPercorso + " (" + Math.round((1.0 * contaPositivi)/lunghPercorso*100) + "%)";
              k++; 
              i = i + 5;
          
          }
        
        return arr;
    }
    

    
    /**
     * Metodo che restituisce il punteggio massimo parziale
     *
     * @return     il punteggio massimo
     */
    public int maxPunteggio()
    {
        Sequenza a = sequenza1();
        Sequenza b = sequenza2();
        int n = a.lunghezza();
        int m = b.lunghezza();
        int pmax;
        // cerca punteggio massimo nell'ultima riga e nell'ultima colonna
        pmax = s[n][0];
        int q = 1; // indice di scorrimento
        while (q < m+1)
           {
             if (s[n][q] > pmax)
                {pmax = s[n][q];}
             q++;           
           }
        q = 0;
        while (q < n)  // non n+1 per non considerare due volte s[n][m]
           {
             if (s[q][m] > pmax)
                {pmax = s[q][m];}
             q++;           
           }        
        return pmax;
    }
    
       
     // metodo ricorsivo ausiliario
     private void risaliPercorso(int i, int j, Vector v, String percorso)
     {
        int z = p[i][j];  //variabile di appoggio
        if ((i==0) || (j==0)) 
            // fine del percorso corrente
           { 
               // devo sincronizzare l'inizio delle due sequenze
               while (i>0)
               {
                  percorso = percorso+"4";
                  i--;
               }
               while (j>0)
               {
                  percorso = percorso+"1";
                  j--;
               }
               // aggiungo il percorso
               v.addElement(percorso);
           }
        else
           {
              if (z == 1)
                 {
                   risaliPercorso(i, j-1, v, percorso+"1");
                 }
              else if (z == 2)
                 {
                   risaliPercorso(i-1, j-1, v, percorso+"2");
                 }
              else if (z == 4)
                 {
                   risaliPercorso(i-1, j, v, percorso+"4");
                 }
              else if (z == 3)
                {
                   // doppio percorso
                   risaliPercorso(i, j-1, v, percorso + "1");
                   String nuovo1 = percorso;
                   risaliPercorso(i-1, j-1, v, nuovo1 + "2");                   
                 }
              else if (z == 5)
                 {
                   // doppio percorso
                   risaliPercorso(i, j-1, v, percorso + "1");
                   String nuovo2 = percorso;
                   risaliPercorso(i-1, j, v, nuovo2 + "4");                   
                 }
              else if (z == 6)
                 {
                   // doppio percorso
                   risaliPercorso(i-1, j, v, percorso + "4");
                   String nuovo3 = percorso;
                   risaliPercorso(i-1, j-1, v, nuovo3 + "2");                   
                 }
              else if (z == 7)
                 {
                   // triplo percorso
                   risaliPercorso(i, j-1, v, percorso + "1");
                   String nuovo4 = percorso;
                   risaliPercorso(i-1, j-1, v, nuovo4 + "2");
                   String nuovo5 = percorso;
                   risaliPercorso(i-1, j, v, nuovo5 + "4");                   
                 }
           }       
     }
   
}
